import pandas as pd
import geopandas as gpd


def addRegressorWeightedAverage(
    df: gpd.GeoDataFrame, regressor_df: gpd.GeoDataFrame, discr_id_col: str = "h3_index"
) -> gpd.GeoDataFrame:
    """
    Apply parameters from regressor database to geographic discretization as a
    weighted average of the areas of intersection.
    Parameters
        df : gpd.GeoDataFrame - geodataframe containing the final geographic
        discretization and a columns 'discr_id' containing unique IDs.
        regressor_df : gpd.GeoDataFrame - geodataframe containing the desired
        regressors and geometries defining geographic boundaries.
        discr_id_col : str - ID column from df
    Returns
        overlay_df : gpd.GeoDataFrame - geodataframe with same structure as df
        with new columns of applied regressors.
    """

    # merge calculating area overlays
    overlay_df = gpd.overlay(df, regressor_df, how="intersection").rename(
        {"geometry": "geo_overlay"}, axis=1
    )
    # merge with df to get original discretization
    overlay_df = pd.merge(overlay_df, df, on=discr_id_col, how="outer").rename(
        {"geometry": "geometry_discr"}, axis=1
    )

    # complete area zero for non overlay dicretizations
    overlay_df["geo_overlay_area"] = overlay_df["geo_overlay"].area.copy()
    overlay_df.loc[overlay_df["geo_overlay"].isna(), "geo_overlay_area"] = 0

    # calculate percentage of areas on each overlay
    overlay_df["overlay_area_percentage"] = (
        overlay_df["geo_overlay_area"] / overlay_df["geometry_discr"].area
    )

    # apply regressor columns multiplying merged values with
    # calculated percentages
    regressor_cols = [
        col for col in regressor_df.columns if not (col in ["geometry", "regr_id"])
    ]
    for regressor_col in regressor_cols:
        overlay_df[regressor_col] *= overlay_df["overlay_area_percentage"]

    # sum weithed partials and recover original columns
    overlay_df = overlay_df.groupby([discr_id_col])[regressor_cols].sum()
    overlay_df = pd.merge(df, overlay_df.reset_index(), on=discr_id_col, how="left")

    return overlay_df


def addRegressorUniformDistribution(
    df: gpd.GeoDataFrame,
    regressor_df: gpd.GeoDataFrame,
    discr_id_col: str = "h3_index",
    type_geo_variable: str = "feature",
) -> gpd.GeoDataFrame:
    """
    Apply parameters from regressor database to geographic discretization as a
    uniform distribution of parameters in regressors area.
    Parameters
        df : gpd.GeoDataFrame - geodataframe containing the final geographic
        discretization and a columns 'discr_id' containing unique IDs.
        regressor_df : gpd.GeoDataFrame - geodataframe containing the desired
        regressors and geometries defining geographic boundaries.
        discr_id_col : str - discretization ID column from df
        type_geo_variable: str - "feature" or "area". specifies if we want to multiply the value of the feature by the
        intersection percentage or just get its area.
    Returns
        overlay_df : gpd.GeoDataFrame - geodataframe with same structure as df
        with new columns of applied regressors.
    """

    # create regressors ID column
    regressor_df["regr_id"] = list(range(len(regressor_df)))

    # merge calculating area overlays
    overlay_df = gpd.overlay(df, regressor_df, how="intersection").rename(
        {"geometry": "geo_overlay"}, axis=1
    )
    # merge with df to get original discretization
    overlay_df = pd.merge(
        overlay_df, regressor_df[["regr_id", "geometry"]], on="regr_id", how="left"
    ).rename({"geometry": "geometry_regr"}, axis=1)

    # calculate percentage of regressor areas on each overlay
    areas = overlay_df["geometry_regr"].area.apply(
        lambda x: 10**-4 if x < 10**-4 else x
    )
    overlay_df["overlay_area_percentage"] = overlay_df["geo_overlay"].area / areas

    # apply regressor columns multiplying merged values with
    # calculated percentages
    regressor_cols = [
        col for col in regressor_df.columns if not (col in ["geometry", "regr_id"])
    ]
    for regressor_col in regressor_cols:
        if type_geo_variable == "feature":
            overlay_df[regressor_col] *= overlay_df["overlay_area_percentage"]
        elif type_geo_variable == "area":
            overlay_df[regressor_col] *= overlay_df["geo_overlay"].area
            overlay_df[regressor_col] /= 10**6

    # sum weithed partials and recover original columns
    overlay_df = overlay_df.groupby([discr_id_col])[regressor_cols].sum()

    overlay_df = pd.merge(df, overlay_df.reset_index(), on=discr_id_col, how="left")

    # If a region in D1 doesn't have intersection with regions in D2, its regressor_cols values may be nan
    # In this case we're replacing them with zero
    for regressor_col in regressor_cols:
        overlay_df[regressor_col] = overlay_df[regressor_col].fillna(0)

    return overlay_df



import pandas as pd
import geopandas as gpd
import numpy as np

import matplotlib.pyplot as plt

from shapely.ops import unary_union
# from scipy.spatial import Voronoi
from shapely.geometry import Polygon, MultiPolygon, Point

from geovoronoi import voronoi_regions_from_coords, points_to_coords



def get_voronoi_regions(voro_points: gpd.GeoDataFrame, borders: gpd.GeoDataFrame):
    bases = voro_points[["geometry"]].copy()
    max_borders = borders[["geometry"]].copy()

    bases["id"] = list(range(len(bases)))
    if bases.crs is None:
        bases = bases.set_crs("epsg:4326")
    else:
        bases = bases.to_crs("epsg:4326")
        
    if max_borders.crs is None:
        max_borders = max_borders.set_crs(epsg=3395)
    else:
        max_borders = max_borders.to_crs(epsg=3395)

    bases_proj = bases.to_crs(max_borders.crs)


    boundary_shape = unary_union(max_borders.geometry)
    coords = points_to_coords(bases_proj.geometry)

    # gc = voronoi_polygons(bases["geometry"], extend_to=max_borders["geometry"])
    poly_shapes, pts, unassigned = voronoi_regions_from_coords(coords, boundary_shape, return_unassigned_points=True,per_geom=False)

    items = poly_shapes.items()
    bases_poly_map = {}
    for key,poly in poly_shapes.items():
        bases_poly_map[key] = []
        for i,row in bases_proj.iterrows():
            if poly.contains(row["geometry"]):
                bases_poly_map[key].append(i)

    voros = pd.DataFrame()
    voros["id"] = [x[0] for x in items]
    voros["geometry"] = [x[1] for x in items]
    voros["id"].replace(bases_poly_map,inplace=True)
    voros = voros.sort_values(by="id")


    voros = gpd.GeoDataFrame(voros, geometry="geometry")
    bases = bases.merge(voros, left_on="id", right_on="id")

    bases = bases.rename(columns={"geometry_y": "geometry"})
    bases = gpd.GeoDataFrame(bases[["geometry"]].copy(), geometry="geometry")
    bases = bases.set_crs(epsg=3395)
    bases = bases.to_crs("epsg:4326")

    return bases


def distance(p1, p2):
    R = 6370
    pi = np.pi
    lat1 = p1[0]
    long1 = p1[1]
    lat2 = p2[0]
    long2 = p2[1]
    cos = np.cos
    sin = np.sin
    asin = np.arcsin
    point1=(R*cos((pi/180)*lat1)*cos((pi/180)*long1), R*cos((pi/180)*lat1)*sin((pi/180)*long1), R*sin((pi/180)*lat1))
    point2=(R*cos((pi/180)*lat2)*cos((pi/180)*long2), R*cos((pi/180)*lat2)*sin((pi/180)*long2), R*sin((pi/180)*lat2))
    d=np.linalg.norm(np.array(point1)-np.array(point2))
    dearth = 2*R*asin(d/(2*R))


    return dearth


import numpy as np
import pandas as pd
import geopandas as gpd

from typing import List
from shapely.ops import unary_union
from shapely.geometry import Polygon, MultiPolygon, Point

from geovoronoi.plotting import subplot_for_map, plot_voronoi_polys_with_points_in_area
from geovoronoi import voronoi_regions_from_coords, points_to_coords

from .time_discretization_utils import calculate_seasonality, apply_custom_time_events
from .geo_discretization_utils import get_voronoi_regions, distance
from .squares import rectangle_discretization
from .h3_utils import generate_H3_discretization
from .add_regressors import addRegressorUniformDistribution, addRegressorWeightedAverage


class DataAggregator:

    def __init__(self, crs: str = "epsg:4326"):
        """
        Class initialization.

        Parameters
        ----------
        crs : The value can be any accepted by pyproj.CRS.from_user_input()
            Coordinate Reference System to be assured and applied.
        """
        self.crs = crs
        self.time_indexes = []
        self.geo_index = "gdiscr"
        self.events_features = []
        self.geo_features = []

        self.max_borders = None
        self.events_data = None
        self.geo_discretization = None

    def add_max_borders(self, data: gpd.GeoDataFrame = None, method: str = None):
        """
        Adds limit borders to internal map. Will be used on multiple
        methods to limit events and discretization reach.

        Parameters
        ----------
        data : geopandas.GeoDataFrame
            GeoDataFrame containing Polygon representing limit borders

        method : str
            Method of estimating borders using events data:

                rectangle => Calculates smallest rectangle containing all event points
                convex => Calculates aproximate of smallest convex polygon containing all event points
        """

        if data is not None:
            # If no CRS set, will assume class set CRS
            if data.crs is None:
                print(
                    f"""Limit borders data is not set with CRS information. Will assume "{self.crs}"."""
                )
                self.max_borders = data

            # If CRS identified, set new one if needed
            elif data.crs != self.crs:
                self.max_borders = data.to_crs(self.crs)

            else:
                self.max_borders = data

        elif method == "rectangle":

            # if no events data was passed, raise error
            if not (hasattr(self, "events_data")):
                raise AttributeError(
                    "Please inform events data using `add_events_data` method."
                )

            # get limits and compute rectangle
            minx, miny, maxx, maxy = self.events_data.geometry.total_bounds
            pol_max_borders = Polygon(
                [(minx, miny), (minx, maxy), (maxx, maxy), (maxx, miny)]
            )
            # replace self.max_borders
            self.max_borders = gpd.GeoDataFrame()
            self.max_borders["geometry"] = [pol_max_borders]
            self.max_borders = self.max_borders.set_crs(self.crs)

        elif method == "convex":

            # if no events data was passed, raise error
            if not (hasattr(self, "events_data")):
                raise AttributeError(
                    "Please inform events data using `add_events_data` method."
                )

            x_std = self.events_data.geometry.x.std() * 0.1
            y_std = self.events_data.geometry.y.std() * 0.1
            # concatenate list of mini event-polygons
            d_polys = self.events_data.drop(
                self.events_data.loc[self.events_data.geometry.x.isna()].index
            ).geometry.apply(
                lambda pt: (
                    Polygon(
                        [
                            (pt.x - x_std, pt.y - y_std),
                            (pt.x - x_std, pt.y + y_std),
                            (pt.x + x_std, pt.y + y_std),
                            (pt.x + x_std, pt.y - y_std),
                        ]
                    )
                    if pt.x >= -np.inf
                    else pt
                )
            )
            # replace self.max_borders
            self.max_borders = gpd.GeoDataFrame()
            self.max_borders["geometry"] = [
                MultiPolygon(list(d_polys.values)).convex_hull
            ]
            self.max_borders = self.max_borders.set_crs(self.crs)

        else:
            raise ValueError("Please inform `data` or `method` parameter.")

    def add_events_data(
        self,
        events_data: pd.DataFrame,
        datetime_col: str = None,
        lat_col: str = "lat",
        lon_col: str = "lon",
        feature_cols: List[str] = [],
        datetime_format: str = None,
    ):
        """
        Processes and save events registers.

        Parameters
        ----------
        events_data : pandas.DataFrame, geopandas.GeoDataFrame
            DataFrame containing latitude longitude information
            -> If pandas dataframe, should contain two
            columns (lat_col, lon_col) with latitude and longitude
            coordinates
            -> If geopandas geodataframe, should contain "geometry" column

        datetime_col : str
            Column containing time information from events
            Can be not informed if not applicable

        lat_col : str
            Column containing latitude from events

        lon_col : str
            Column containing longitude from events

        feature_cols : List[str]
            Column(s) containing adicional information from events
            that will be considered in discretization process

        datetime_format : str
            The strftime to parse time, e.g. "%d/%m/%Y".
            See https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
            for reference
        """
        # configure geodataframe
        # if pandas dataframe type, will create geometry columns
        if type(events_data) == pd.DataFrame:
            events_data["geometry"] = gpd.points_from_xy(
                events_data[lon_col], events_data[lat_col]
            )
            self.events_data = (
                gpd.GeoDataFrame(events_data).set_crs("epsg:4326").to_crs(self.crs)
            )
        # if geopandas object, will assure correct CRS
        else:
            if events_data.crs is None:
                self.events_data = events_data.set_crs(self.crs)
            else:
                self.events_data = events_data.to_crs(self.crs)

        # transform datetime_col data to datetime
        if datetime_col is not None:
            ts = self.events_data[datetime_col]
            if ts.dtype == "O":
                self.events_data["ts"] = pd.to_datetime(ts, format=datetime_format)
            else:
                self.events_data["ts"] = ts.copy()
            self.events_data.sort_values("ts", ascending=True, inplace=True)

        # filter only necessary cols
        self.events_features += feature_cols
        special_cols = ["geometry", "ts"] if datetime_col is not None else ["geometry"]
        self.events_data = self.events_data[feature_cols + special_cols].copy()

    def add_time_discretization(
        self,
        seasonality_type: str,
        window: int = 1,
        frequency: int = 1,
        column_name: str = None,
    ):
        """
        Applies seasonality index as time discretization new column.

        Parameters
        ----------
        seasonality_type: str or DataFrame
            If string, represents seasonality frequency type
                Y -> year
                M -> month
                W -> week
                D -> day
                H -> hour
                m -> minute
                S -> second

            If dataframe, represents all necessary information for each event.

        window: (List[int], int)
            Seasonality window. Can be unique, passed as an int, or variable,
            passed as a list of int.

        frequency: int
            Seasonality frequency.

            eg: If seasonality is between 12 months in a year:
                seasonality_type = 'Y'
                window = 1
                frequency = 12

                If time discretization must consider the first 12 hours in a day as first index,
                next 8 hours as second and last 4 hours as third:
                seasonality_type = 'D'
                window = [12, 8, 4]
                frequency = 24
        """
        # throw error if events_data is None
        if self.events_data is None:
            raise AttributeError(
                "Please inform events data using `add_events_data` method."
            )

        # get new column name
        if column_name is None:
            i_col = 1
            while True:
                t_col = f"tdiscr_{i_col}"
                if t_col in self.events_data.columns:
                    i_col += 1
                else:
                    break
        else:
            t_col = column_name

        # if dataframe passed, apply custom events time discretization
        if type(seasonality_type) == pd.DataFrame:
            self.time_indexes.append(t_col)
            self.events_data[t_col] = apply_custom_time_events(
                ts=self.events_data["ts"], time_disc_df=seasonality_type, nan_idx=0
            )

        else:
            # throw error if frequency is not compatible with window
            if (type(window) is int) and (frequency % window != 0):
                raise ValueError(
                    "Parameter `frequency` must be a multiple of `window`."
                )
            if type(window) is list and sum(window) != frequency != 0:
                raise ValueError("Parameter `frequency` must be equal sum of `window`.")

            # add new column
            self.time_indexes.append(t_col)
            self.events_data[t_col] = calculate_seasonality(
                ts=self.events_data["ts"],
                seasonality_type=seasonality_type,
                window=window,
                frequency=frequency,
            )
        self.events_data = self.events_data.dropna()

    def add_geo_discretization(
        self,
        discr_type: str,
        hex_discr_param: int = None,
        rect_discr_param_x: int = None,
        rect_discr_param_y: int = None,
        custom_data: gpd.GeoDataFrame = None,
    ) -> gpd.GeoDataFrame:
        """
        Calculates geographical discretization and apply index
        to events data.

        Parameters
        ----------
        discr_type : str
            Indicator of method of geographical discretization.

                R => Rectangular discretization. Parameters `rect_discr_param_x`
                and `rect_discr_param_y` are required.

                H => Hexagonal discretization using H3 package. Parameter
                `hex_discr_param` is required.

                C => Custom discretization using user GeoDataFrame. Parameter
                `custom_data` is required.

                G => Custom discretization using graph data. Geolocated nodes
                must be informed in dataframe format `custom_data`.

                V => Voronoi discretization using GeoDataFrame with points.
                `custom_data` must contain the points that will define the voronoi regions.

        hex_discr_param : int
            Granularity level for hexagonal discretization.

        rect_discr_param_x : int
            Granularity level for rectangular horizontal discretization.

        rect_discr_param_y : int
            Granularity level for rectangular vertical discretization.
        """

        # if there is no max borders set throw error
        if self.max_borders is None:
            raise ValueError(
                "Please inform geographical limits using `add_max_borders` method."
            )

        # square discretization
        if discr_type == "R":
            self.geo_discretization = rectangle_discretization(
                gdf=self.max_borders, nx=rect_discr_param_x, ny=rect_discr_param_y
            )

        # hexagonal discretization
        elif discr_type == "H":
            full_area = self.max_borders.copy()  # keep original
            full_area.geometry = (
                full_area.geometry.convex_hull
            )  # apply convex hull so no area is cutted out
            self.geo_discretization = generate_H3_discretization(
                self.max_borders, hex_discr_param
            )

            # cut hexagons borders that dont belong to original shape
            limited_discretization = (
                gpd.overlay(
                    self.geo_discretization[["id", "geometry"]],
                    self.max_borders[["geometry"]],
                    how="intersection",
                )
                .dissolve(by="id")
                .reset_index()
            )
            self.geo_discretization = pd.merge(
                self.geo_discretization.drop("geometry", axis=1),
                limited_discretization,
                on="id",
                how="right",
            )
            # return type to geoDataFrame
            self.geo_discretization = gpd.GeoDataFrame(self.geo_discretization)

            # correct polygons index, ordering in neighbors list
            corr_index_dict = (
                self.geo_discretization.reset_index().set_index("id")["index"].to_dict()
            )
            neighbors_list = list(self.geo_discretization["neighbors"])
            for ind, n_list in enumerate(neighbors_list):
                new_list = []
                for n in n_list:
                    if corr_index_dict.get(n):
                        new_list.append(corr_index_dict[n])
                neighbors_list[ind] = new_list
            self.geo_discretization["neighbors"] = neighbors_list
            self.geo_discretization.drop("id", axis=1, inplace=True)
            self.geo_discretization.reset_index(inplace=True)
            self.geo_discretization.rename({"index": "id"}, axis=1, inplace=True)

        elif discr_type == "C":
            # check for custom_data
            if custom_data is None:
                raise ValueError(
                    "Please inform `custom_data` to be used as geographical discretization."
                )
            else:
                self.geo_discretization = custom_data[["geometry"]].copy()
                self.geo_discretization["id"] = list(range(len(custom_data)))

                # If no CRS set, will assume class set CRS
                if self.geo_discretization.crs is None:
                    print(
                        f'Custom data is not set with CRS information. Will assume "{self.crs}".'
                    )
                    self.geo_discretization = self.geo_discretization.set_crs(self.crs)

                # If CRS indetified, set new one if needed
                elif self.geo_discretization.crs != self.crs:
                    self.geo_discretization = self.geo_discretization.to_crs(self.crs)

            # compute neighbors
            neighbors = []
            for _, row in self.geo_discretization.iterrows():
                row_neighbors = list(
                    self.geo_discretization[
                        ~self.geo_discretization.geometry.disjoint(row.geometry)
                    ]["id"]
                )
                row_neighbors = [n for n in row_neighbors if n != row["id"]]
                neighbors.append(row_neighbors)
            self.geo_discretization["neighbors"] = neighbors

        elif discr_type == "G":
            # check for custom_data
            if custom_data is None:
                raise ValueError(
                    "Please inform `custom_data` to be used as grpah discretization."
                )
            else:
                self.geo_discretization = custom_data[["geometry"]].copy()
                self.geo_discretization["id"] = list(range(len(custom_data)))

                # If no CRS set, will assume class set CRS
                if self.geo_discretization.crs is None:
                    print(
                        f'Custom data is not set with CRS information. Will assume "{self.crs}".'
                    )
                    self.geo_discretization = self.geo_discretization.set_crs(self.crs)

                # If CRS indetified, set new one if needed
                elif self.geo_discretization.crs != self.crs:
                    self.geo_discretization = self.geo_discretization.to_crs(self.crs)

            # join events with nodes using nearest node
            self.events_data = (
                gpd.sjoin_nearest(
                    self.events_data.to_crs("epsg:29193"),
                    self.geo_discretization.to_crs("epsg:29193")[["id", "geometry"]],
                    how="left",
                )
                .to_crs(self.crs)
                .drop("index_right", axis=1)
                .rename(columns={"id": "node_id"})
            )

            # create map from node_id to gdiscr index
            node_idx_dict = dict(
                enumerate(self.events_data["node_id"].dropna().unique())
            )
            node_idx_dict = {v: k for k, v in node_idx_dict.items()}
            self.events_data["gdiscr"] = self.events_data["node_id"].map(node_idx_dict)
            self.events_data.drop("node_id", axis=1, inplace=True)

            # drop duplicated events
            self.events_data = (
                self.events_data.reset_index()
                .drop_duplicates(subset=["index"], keep="first")
                .set_index("index")
            )

            # not consider events outside max borders
            self.events_data.loc[
                ~self.events_data["geometry"]
                .dropna()
                .within(self.max_borders["geometry"].values[0]),
                "gdiscr",
            ] = pd.NA

            # column gdiscr is alread copmuted so function must be terminated
            return None
        elif discr_type == "V":
            if custom_data is None:
                raise ValueError(
                    "Please inform `custom_data` to be used as geographical discretization."
                )
            else:
                disc = get_voronoi_regions(custom_data, self.max_borders)
                self.geo_discretization = disc[["geometry"]].copy()
                self.geo_discretization["id"] = list(range(len(custom_data)))
                # self.geo_discretization = custom_data[['geometry']].copy()
                # self.geo_discretization['id'] = list(range(len(custom_data)))

                # If no CRS set, will assume class set CRS
                if self.geo_discretization.crs is None:
                    print(
                        f'Custom data is not set with CRS information. Will assume "{self.crs}".'
                    )
                    self.geo_discretization = self.geo_discretization.set_crs(self.crs)

                # If CRS identified, set new one if needed
                elif self.geo_discretization.crs != self.crs:
                    self.geo_discretization = self.geo_discretization.to_crs(self.crs)

            # compute neighbors
            neighbors = []
            for _, row in self.geo_discretization.iterrows():
                row_neighbors = list(
                    self.geo_discretization[
                        ~self.geo_discretization.geometry.disjoint(row.geometry)
                    ]["id"]
                )
                row_neighbors = [n for n in row_neighbors if n != row["id"]]
                neighbors.append(row_neighbors)
            self.geo_discretization["neighbors"] = neighbors
        else:
            raise ValueError(f"Invalid `discr_type` value {discr_type}.")

        # fills center_lat and center_lon (using projected crs for distance precision)
        centroids = self.geo_discretization.to_crs("epsg:4088").geometry.centroid
        self.geo_discretization["center_lat"] = centroids.x
        self.geo_discretization["center_lon"] = centroids.y

        # apply index to events data
        self.events_data = (
            gpd.sjoin(
                self.events_data.drop("gdiscr", axis=1, errors="ignore"),
                self.geo_discretization[["geometry", "id"]],
                how="left",
                predicate="within",
            )
            .drop("index_right", axis=1)
            .rename({"id": "gdiscr"}, axis=1)
        )

        self.events_data = self.events_data.dropna()

    def add_geo_variable(
        self, data: gpd.GeoDataFrame, type_geo_variable: str = "feature"
    ):
        """
        Merge information from external geographical data with computed
        geographical discretization.

        Parameters
        ----------
        data: geopandas.GeoDataFrame
            Geodataframe with desired variables.
        """

        if type_geo_variable != "feature" and type_geo_variable != "area":
            raise ValueError(
                'Parameter `type_geo_variable` must be "feature" or "area".'
            )

        # If no CRS set, will assume class set CRS
        if data.crs is None:
            print(
                f'Regressor data is not set with CRS information. Will assume "{self.crs}".'
            )
            regr_data = data.set_crs(self.crs)

        # If CRS indetified, set new one if needed
        else:
            regr_data = data.to_crs(self.crs)
        # add features to geo_features list
        self.geo_features += list(data.drop(["geometry"], axis=1).columns)

        # Calculate intersection between regressor data and geo discretization.
        # For math precision, if current crs is not projected,
        # will transform into proper (projected) CRS before using area metric
        # to calcultate intersections
        geo_disc = self.geo_discretization[["id", "geometry"]].copy()
        if not (self.geo_discretization.crs.is_projected):
            geo_disc = geo_disc.to_crs("epsg:4088")
            regr_data = regr_data.to_crs("epsg:4088")

        regr_intersection = addRegressorUniformDistribution(
            geo_disc, regr_data, discr_id_col="id", type_geo_variable=type_geo_variable
        )
        # regr_intersection = addRegressorWeightedAverage(
        #     geo_disc,
        #     regr_data,
        #     discr_id_col='id'
        # )
        self.geo_discretization = pd.merge(
            self.geo_discretization,
            regr_intersection.drop("geometry", axis=1),
            on="id",
            how="left",
        )

    def write_files(self, geo_discretization_path: str, events_data_path: str):
        """Write discretization DataFrames into files provided by geo_discretization_path and events_data_path.
            Files are saved as csv.
        Parameters
        ----------
        geo_discretization_path: str
            path containing location where to save the geo_discretization DataFrame.
        events_data_path: str
            path containing location where to save the events DataFrame.
        """
        self.geo_discretization.to_csv(geo_discretization_path, index=False)
        self.events_data.to_csv(events_data_path, index=False)

    def get_intersection(self, df_geo1: gpd.GeoDataFrame, df_geo2: gpd.GeoDataFrame):
        """
            Calculates the intersection between GeoDataFrames df_geo1 and df_geo2
        Parameters
        ----------
        df_geo1: gpd.GeoDataFrame
        df_geo2: gpd.GeoDataFrame
        """
        crs1 = df_geo1.crs
        crs2 = df_geo2.crs
        df_geo1 = df_geo1.assign(row_number_1=range(len(df_geo1)))
        df_geo2 = df_geo2.assign(row_number_2=range(len(df_geo2)))
        df_geo1 = df_geo1.to_crs("epsg:32723")
        df_geo2 = df_geo2.to_crs("epsg:32723")
        intersec = df_geo1.overlay(df_geo2, how="intersection")

        A = np.zeros((len(df_geo1), len(df_geo2)))
        intersec = intersec.to_crs("epsg:32723")
        for k, row in intersec.iterrows():
            i = int(row["row_number_1"])
            j = int(row["row_number_2"])
            A[i, j] += row["geometry"].area / 10**6

        df_geo1.drop(columns=["row_number_1"])
        df_geo2.drop(columns=["row_number_2"])
        df_geo1 = df_geo1.to_crs(crs1)
        df_geo2 = df_geo1.to_crs(crs2)
        return A

    def get_events_aggregated(self):
        """Returns np.array of lists with number of events given by time indexes, geodiscretization index and features indexes for each observation."""
        limits = []
        for time_index in self.time_indexes:
            limits.append(np.max(self.events_data[time_index]) + 1)
        limits.append(int(np.max(self.events_data["gdiscr"]) + 1))
        for feature in self.events_features:
            limits.append(np.max(self.events_data[feature]) + 1)

        samples = np.empty(tuple(limits), dtype=object)
        for index, val in np.ndenumerate(samples):
            samples[index] = []

        i = 0
        self.events_data.dropna()
        while i < len(self.events_data):
            row = self.events_data.iloc[i]
            times_i = [row[time_index] for time_index in self.time_indexes]
            index_i = (
                times_i
                + [int(row["gdiscr"])]
                + [row[feature] for feature in self.events_features]
            )
            count = 1
            for j in range(i + 1, len(self.events_data)):
                row_j = self.events_data.iloc[j]
                times_j = [row_j[time_index] for time_index in self.time_indexes]
                index_j = (
                    times_j
                    + [int(row_j["gdiscr"])]
                    + [row_j[feature] for feature in self.events_features]
                )

                if index_i == index_j:
                    count += 1
                elif times_i != times_j:
                    break
            samples[tuple(index_i)].append(count)
            i += count

        return samples

    def write_info(self, obs_index_column, path="info.dat"):
        """Write file with informations about given discretization.

        Args:
            obs_index_column (str): time index that defines the number of observations.
            path (str, optional): path of the written file. Defaults to "info.dat".
        """
        info_file = open(path, "w")
        for time_index in self.time_indexes:
            info_file.write(f"{np.max(self.events_data[time_index]) + 1} ")
        info_file.write(f"{int(np.max(self.events_data['gdiscr']) + 1)} ")
        for feature in self.events_features:
            info_file.write(f"{np.max(self.events_data[feature]) + 1} ")
        info_file.write(f"{len(self.geo_features)} ")
        nb_holidays = 0
        info_file.write(f"{nb_holidays}\n")
        events = self.events_data
        nb_obs_column_indexes = np.max(events[obs_index_column]) + 1
        obs_count = [0] * nb_obs_column_indexes
        for d in range(nb_obs_column_indexes):
            filtered_events = events.loc[events[obs_index_column] == d].copy()
            filtered_events["date_only"] = pd.to_datetime(filtered_events["ts"]).copy()
            date_only = filtered_events["date_only"].dt.date
            obs_count[d] = len(date_only.unique())
            info_file.write(f"{obs_count[d]} ")
        # print(f"obs_count = {obs_count}")
        # info_file.write(obs_count)
        info_file.write("\nEND")
        info_file.close()

    def write_arrivals(self, path="arrivals.dat"):
        """Write events data, aggregated by time indexes, geo index and features.

        Args:
            path (str, optional): path of the written file. Defaults to "arrivals.dat".
        """
        samples = self.get_events_aggregated()
        arrivals_file = open(path, "w")
        # arrivals_file.write("%s\n" % (" ".join([str(x) for x in samples.shape])))
        for index, sample in np.ndenumerate(samples):
            for i, val in enumerate(sample):
                line = "%s %d %d\n" % (" ".join([str(x) for x in index]), i, val)
                arrivals_file.write(line)
        arrivals_file.write("END")
        arrivals_file.close()

    def write_regions(self, path="neighbors.dat"):
        """Write file with description of discretization subregions.
        Description contains for each subregion:
            id,
            centroid coordinates,
            geo features,
            pairs of neighbor subregion ids and distances (km).

        Args:
            path (str, optional): path of the written file. Defaults to "neighbors.dat".
        """
        centers = self.geo_discretization.geometry.to_crs("epsg:29193").centroid.to_crs(
            self.geo_discretization.crs
        )
        coords = centers.apply(lambda x: x.representative_point().coords[:][0])
        neighbors_file = open(path, "w")
        # neighbors_file.write("%d %d\n" % (len(centers), len(self.geo_features)))
        # print(f"Writing geo_features {self.geo_features}")
        for i, row in self.geo_discretization.iterrows():
            id_region = row["id"]
            lat = coords[id_region][1]
            lon = coords[id_region][0]
            neighbors = row["neighbors"]
            features = [row[feature] for feature in self.geo_features]
            neighbors_file.write(
                "%d %.6f %.6f 0 %s "
                % (id_region, lat, lon, " ".join([str(x) for x in features]))
            )
            for neighbor in neighbors:
                row_neighbor = self.geo_discretization[
                    self.geo_discretization["id"] == neighbor
                ].iloc[0]
                id_neighbor = row_neighbor["id"]
                lat2, lon2 = coords[id_neighbor][1], coords[id_neighbor][0]
                d = distance((lat, lon), (lat2, lon2))
                neighbors_file.write(" %d %.3f" % (row_neighbor["id"], d))
            neighbors_file.write("\n")
        neighbors_file.write("END")
        neighbors_file.close()

    def plot_discretization(self, to_file=None):
        """Plot the current discretization. The plot contains: events as points, max_borders and subregion borders and ids.

        Args:
            to_file (string, optional): path indicating where to save the discretization figure. 

        Returns:
            _type_: _description_
        """
        import matplotlib.pyplot as plt

        self.geo_discretization["center"] = self.geo_discretization.geometry.to_crs(
            "epsg:29193"
        ).centroid.to_crs(self.geo_discretization.crs)
        self.geo_discretization["coords"] = self.geo_discretization["center"].apply(
            lambda x: x.representative_point().coords[:][0]
        )

        fig, ax = plt.subplots(figsize=(24, 16))
        self.geo_discretization.boundary.plot(ax=ax, aspect=1)
        self.events_data.dropna(subset=["gdiscr"]).plot(
            markersize=5, color="red", ax=ax
        )
        for _, row in self.geo_discretization.iterrows():
            plt.annotate(
                text=row["id"],
                xy=row["coords"],
                horizontalalignment="center",
                verticalalignment="center",
                color="black",
                fontsize=11,
            )
        if to_file != None:
            plt.savefig(to_file, bbox_inches="tight")
            return f"saved discretization plot at {to_file}"
        else:
            plt.show()
            return None



import h3
from geopandas import GeoDataFrame
from shapely.geometry import Polygon, Point


# helper function: gets geojson like (H3 expects as input) from shapely polygon
def polygon_to_geojson(polygon: Polygon):
    """
    Auxiliar function: Gets geojson like (H3 expect as input) from
    Shapely Polygon object
    """
    temp_coord_list = list(polygon.exterior.coords)
    # we have to manually convert from a list of list of tuples
    # to a list of list of lists (and also invert lat, long ordering)
    coords = [[lat, long] for (long, lat) in temp_coord_list]
    # we dont want the loop around here
    coords.pop()
    # generate geojson
    geoJson = {"type": "Polygon", "coordinates": [coords]}

    return geoJson


def generate_H3_discretization(gdf: GeoDataFrame, resolution: int = 7):
    """
    Generate a hexagonal discretization of the area using Uber's H3 library
    and returns a new GeoDataFrame with it.

    Parameters
    ----------
    param gdf: GeoDataFrame
        Original geodataframe whose boundaries will be considered to generate the discretization
    param resolution: int
        Desired resolution level passed to the H3 library. A bigger number means smaller hexagons.
        Valid range : [0,15]

    Returns
    -------
    return: GeoDataFrame
        New GeoDataFrame containing the H3 hexagons that approximately cover the original region
    """

    if resolution < 0 or resolution > 15:
        raise ValueError(
            "Hex resolution must be in range [0, 15]. Got {}".format(resolution)
        )

    # following code assumes crs epsg = 4326 to interface with H3
    h3_gdf = gdf.to_crs(epsg=4326)

    # loop through all observation in the geoseries
    # For every polygon there, compute using h3 the a list of h3 indexes
    # and add it to the hex_indexes structure, used later to construct a new goodataframe

    # this procedure "flattens" the original geoseries, ie, multipolygon's are
    # treated as a sequence of polygon without differentiating them in any way

    hex_indexes = set()
    for observation in h3_gdf.geometry:
        if observation.geom_type == "MultiPolygon":
            for polygon in observation.geoms:
                geoJson = polygon_to_geojson(polygon)
                # h3.polyfill is the important method in the H3 library
                # that does the heavy work of finding a good hex-cover
                hex_indexes.update(h3.polyfill(geoJson, resolution))
        elif observation.geom_type == "Polygon":
            geoJson = polygon_to_geojson(observation)
            hex_indexes.update(h3.polyfill(geoJson, resolution))
        else:
            raise ValueError(
                "GeoDataFrame's geometry is limited to either polygon of multi-polygon. Got {}".format(
                    observation.geom_type
                )
            )

    # we have the H3 indexes in hex_indexes
    # Now we just need to transform them into an geodataframe with any relevant info we may need
    hex_indexes = list(hex_indexes)
    polygons = []
    pol_areas = []
    neighbors = []
    center_points = []
    c_neighbors = [[], [], [], [], [], []]
    for hex in hex_indexes:
        # send a warning if there is a pentagon in the study region!
        if h3.h3_is_pentagon(hex):
            print(
                "A H3 cell in the study region is a pentagon. See H3's documentation for further details."
            )

        # once again: h3 uses 2-tuples for points, but shapely uses 2-lists
        hex_coords_h3 = h3.h3_set_to_multi_polygon([hex], geo_json=False)
        hex_coords_sh = []
        for coords in hex_coords_h3[0]:
            for coord in coords:
                # and in a different (lat,long) order
                hex_coords_sh.append([coord[1], coord[0]])

        pol = Polygon(hex_coords_sh)
        polygons.append(pol)
        # cell area in default km2
        pol_areas.append(h3.cell_area(hex))

        # collects h3 center points. There might be a very slight difference between these points
        # and the ones calculated by shapely based on the polygons!
        lat_long_center = h3.h3_to_geo(hex)
        center_points.append(Point(lat_long_center[1], lat_long_center[0]))

        hex_ring = h3.hex_ring(hex, k=1)
        neighbors.append(
            [
                hex_indexes.index(neighbor)
                for neighbor in hex_ring
                if neighbor in hex_indexes
            ]
        )
        i_n = 0
        i_filled = 0
        for neighbor in hex_ring:
            if neighbor in hex_indexes:
                c_neighbors[i_filled].append(neighbor)
                i_filled += 1
            i_n += 1
        assert (
            i_n == 6
        ), "Cell found that did not have 6 neighbors. Does the study area contain H3 pentagons? Check H3 official documentation for details"
        while i_filled < i_n:
            c_neighbors[i_filled].append(None)
            i_filled += 1

    # is there an other relevant info that could be calculated here?
    temp_dict = {"geometry": polygons, "neighbors": neighbors}

    hex_gdf = (
        GeoDataFrame(temp_dict, crs="EPSG:4326")
        .to_crs(gdf.crs)
        .reset_index()
        .rename({"index": "id"}, axis=1)
    )

    return hex_gdf



import geopandas as gpd
from shapely.geometry import Polygon


def rectangle_discretization(
    gdf: gpd.GeoDataFrame,
    nx: int,
    ny: int,
    neighborhood: int = 8
):
    '''
    Construct a square discretized gdf from the original region. The final
    discretization is intended to be a grid of nx per ny rectangles.
    Also construct neighbors data for each rectangle.

    Please be aware that each cell in the returned GeoDataFrame is not
    necessarily a rectangle: edge regions or 'islands' will fit to their
    original formats.  A data structural consequence of this is that the
    shapes forming the regions are not all Polygons.
    They may be Multi-polygons, or Points / Multipoint depending on the
    original topology.

    Arguments:
        gdf : GeoDataFrame
            The GeoDataFrame describing the original study region
        nx, ny
            Define how many subdivision to use on the discretization,
            on the x and y axis respectively
    '''
    # check for negative discretization sizes
    if nx <= 0 or ny <= 0:
        raise ValueError("Squares discretization sizes must be positive")
    # get bounds
    minx, miny, maxx, maxy = gdf.geometry.total_bounds
    # calculate square sizes
    deltay = (maxy - miny) / ny
    deltax = (maxx - minx) / nx

    # construct 'squares' GeoDataFrame
    # Lots of rectangles covering the entire study region
    squares = []
    indices = []
    neighbors = []
    # valid_index = lambda ix, iy : True if ix >= 0 and iy >= 0 and ix < nx
    # and iy < ny else False
    # index = lambda ix, iy : iy * ny + ix if valid_index(ix, iy) else None

    def get_index(ix, iy):
        '''
        Auxiliary function that gets square list index
        based on position (ix, iy)
        '''
        if ix >= 0 and iy >= 0 and ix < nx and iy < ny:
            return iy * ny + ix
        else:
            return None

    for iy in range(ny):
        for ix in range(nx):
            cx, cy = minx + ix * deltax, miny + iy * deltay
            pol = Polygon([
                (cx, cy),
                (cx + deltax, cy),
                (cx + deltax, cy + deltay),
                (cx, cy + deltay)
            ])
            squares.append(pol)
            indices.append(get_index(ix, iy))

            if neighborhood == 4:
                my_neighbors = [
                    get_index(ix + 1, iy),
                    get_index(ix - 1, iy),
                    get_index(ix, iy + 1),
                    get_index(ix, iy - 1)
                ]
            elif neighborhood == 8:
                my_neighbors = [
                    get_index(ix + 1, iy),
                    get_index(ix - 1, iy),
                    get_index(ix, iy + 1),
                    get_index(ix, iy - 1),
                    get_index(ix + 1, iy + 1),
                    get_index(ix + 1, iy - 1),
                    get_index(ix - 1, iy + 1),
                    get_index(ix - 1, iy - 1)
                ]
            else:
                raise ValueError("Invalid neighborhood type " + neighborhood)
            my_neighbors = [n for n in my_neighbors if n is not None]
            neighbors.append(my_neighbors)

    squares = gpd.GeoSeries(squares)
    squares_gdf = gpd.GeoDataFrame({
        'geometry': squares,
        'myindex': indices,
        'neighbors': neighbors
        }, crs=gdf.crs)

    # calculate the intersection between this squares gdf and
    # the original region
    res_intersection = gpd.overlay(squares_gdf, gdf, how='intersection')

    # When calculating the intersection, some rectangles may have been
    # dropped entirely. Thus, we need to adjust the 'neighbor' column
    # to correctly reflect the new indices. That's why we keep a auxiliary
    # 'myindex' column. We just need to adjust 'myindex' to actual indexes.
    new_neighbors = []
    for i, row in res_intersection.iterrows():
        new_list = [
            res_intersection.index[res_intersection['myindex'] == n]
            for n in row['neighbors']
        ]
        new_list = [item for sublist in new_list for item in sublist]
        new_neighbors.append(new_list)
    res_intersection['neighbors'] = new_neighbors

    res_intersection.drop(['myindex'], axis=1, inplace=True)
    res_intersection.reset_index(inplace=True)
    res_intersection.rename(columns={'index': 'id'}, inplace=True)
    return res_intersection[['id', 'neighbors', 'geometry']]



import numpy as np


class Param(object):
    def __init__(self) -> None:
        self.EPS = 1e-5
        self.sigma = 0.5
        self.gap = 0.01
        self.max_iter = 30
        self.lower_lambda = 1e-6
        self.upper_lambda = 1.0
        self.beta_bar = 2.0
        self.cv_proportion = 0.2


class RegularizedModel(object):
    def __init__(
        self,
        N,
        M,
        durations,
        groups,
        weights,
        alphas,
        distance,
        type_region,
        neighbors,
        param,
    ):
        if len(N.shape) != 3:
            raise ValueError(f"N dimension must be 3: C,R,T, but got {len(N.shape)}")
        if len(M.shape) != 3:
            raise ValueError(f"M dimension must be 3: C,R,T, but got {len(M.shape)}")

        self.C, self.R, self.T = N.shape
        self.nb_observations = N
        self.nb_arrivals = M
        self.durations = durations
        self.alpha = alphas
        self.weights = weights
        self.distance = distance
        self.neighbors = neighbors
        self.type_region = type_region
        self.groups = groups
        self.nb_groups = len(groups)
        self.param = param

        self.which_group = np.zeros((self.T))
        for g in range(self.nb_groups):
            for i in groups[g]:
                self.which_group[i] = g
                
        self.emp_rates_by_class = np.zeros((self.C))
        for c in range(self.C):
            sum_c = 0.0
            for r in range(self.R):
                for t in range(self.T):
                    sum_c += self.nb_arrivals[c,r,t] / self.nb_observations
                    
        

    def f(self, x):
        obj = 0
        nb_observations = self.nb_observations
        nb_arrivals = self.nb_arrivals
        durations = self.durations
        alpha = self.alpha
        distance = self.distance

        for c in range(self.C):
            for r in range(self.R):
                for t in range(self.T):
                    obj += nb_observations[c, r, t] * x[c, r, t] * durations[
                        t
                    ] - nb_arrivals[t] * np.log(x[c, r, t])
                    for s in self.neighbors[r]:
                        obj += (
                            0.5
                            * alpha[r, s]
                            * nb_observations[c, r, t]
                            * nb_observations[c, s, t]
                            * ((x[c, r, t] - x[c, s, t]) ** 2)
                            / distance[r, s]
                        )

        for c in range(self.C):
            for r in range(self.R):
                for grindex in range(self.nb_groups):
                    group = self.groups[grindex]
                    for t in group:
                        for t1 in group:
                            if t != t1:
                                obj += (
                                    0.5
                                    * self.weights[grindex]
                                    * nb_observations[c, r, t]
                                    * nb_observations[c, r, t1]
                                    * (x[c, r, t] - x[c, r, t1]) ** 2
                                )
        return obj

    def gradient(self, x):
        grad = np.zeros(x.shape)

        return grad

    def projection(self, x):
        z = np.zeros(x.shape)
        for c in range(self.C):
                for r in range(self.R):
                    for t in range(self.T):
                        z = max(self.param.lower_lambda, x[c,r,t])
        if not self.param.relax_emp_fix:
            for c in range(self.C):
                sum_c = 0.0
                for r in range(self.R):
                    for t in range(self.T):
                        sum_c += z[c,r,t]*self.durations[t]
                diff = self.emp_rates_by_class[c] - sum_c
                if s
        return z



import numpy as np
import pandas as pd
from datetime import timedelta


def apply_time_frequency(ts: pd.Series, window: str, window_size: int):
    '''
    Auxiliar function that applies multiple time frequencies index.

    Parameters
    ----------
    ts: pandas.Series
        Timeseries that will be tranformed into time frequency indexes

    window: str
        Window frequency.
        Y -> year
        M -> month
        W -> week
        D -> day
        H -> hour
        m -> minute
        s -> second

    window_size: int
        How many windows to aggregate into one index.

    Returns
    -------
    w: pd.Series
        Index series.

    Example
    -------
    >>> ts = pd.to_datetime(pd.Series([
            '2022-01-01', '2022-01-04', '2022-01-10', '2022-12-31'
        ]))
    >>> apply_time_frequency(ts, 'W', 1)
        0     0.0
        1     1.0
        2     2.0
        3    52.0
        dtype: float64
    '''
    # apply window unit

    if window == 'Y':
        w = ts.dt.year - ts.dt.year.min()

    elif window == 'M':
        months = ts.dt.month
        years = ts.dt.year
        min_year = years.min()
        w = 12*(years - min_year) + months - 1

    elif window == 'W':
        min_day = ts.min().replace(hour=0, minute=0, second=0)
        min_day = min_day - timedelta(days=min_day.weekday())
        w = (ts - min_day).dt.total_seconds() // (7*24*60*60)

    elif window == 'D':
        min_day = ts.min().replace(hour=0, minute=0, second=0)
        w = (ts - min_day).dt.total_seconds() // (24*60*60)

    elif window == 'H':
        min_hour = ts.min().replace(hour=0, minute=0, second=0)
        w = (ts - min_hour).dt.total_seconds() // (60*60)

    elif window == 'm':
        min_minute = ts.min().replace(hour=0, minute=0, second=0)
        w = (ts - min_minute).dt.total_seconds() // (60)

    elif window == 's':
        min_second = ts.min().replace(second=0)
        w = (ts - min_second).dt.total_seconds()

    # window_size indicates how many consecutive windows should be aggregated
    return w // window_size


def calculate_seasonality(
    ts: pd.Series,
    seasonality_type: str,
    window: int,
    frequency: int
):
    '''
    Calculates seasonality index from multiple time windows and frequencies.

    Parameters
    ----------
    ts: pandas.Series
        Timeseries to be transformed into seasonality indexes.
        MUST BE sorted.

    seasonality_type: str
        Seasonality frequency type
            Y -> year
            M -> month
            W -> week
            D -> day
            H -> hour
            m -> minute
            S -> second

    window: (int, List[int])
        Seasonality window

    frequency: int
        Seasonality frequency

        eg: If seasonality is between 12 months in a year:
            seasonality_type = 'Y'
            window = 1
            frequency = 12

    Returns
    -------
    season_idx: pandas.Series
        Seasonality indexes

    Example
    -------
    >>> ts = pd.to_datetime(pd.Series([
            '2022-01-01', '2022-01-04', '2022-01-10', '2022-12-31'
        ]))
    >>> calculate_seasonality(ts, 'D', 1, 7)
        0     0
        1     3
        2     2
        3     0
        dtype: float64
    '''
    unitary_seasonality = apply_time_frequency(
        ts,
        seasonality_type,
        1
    )
    # if unique frequency, apply basic rule
    if type(window) is int:
        season_idx = (unitary_seasonality // window) % (frequency // window)

    # if variable frequencies, calculate indexes accordingly
    else:
        season_n = unitary_seasonality % frequency
        season_idx = pd.Series(np.nan, index=season_n.index)

        for i_w, w in enumerate(np.cumsum(window)):
            season_idx.loc[(season_n < w) & season_idx.isna()] = i_w

    return season_idx.astype(int)


def apply_custom_time_events(
    ts: pd.Series,
    time_disc_df: pd.DataFrame,
    nan_idx=0
):

    # list of valid years
    years_list = ts.dt.year.unique()

    # validate column type
    for col in ["start", "end"]:
        if time_disc_df[col].dtype == "O":
            time_disc_df[col] = pd.to_datetime(time_disc_df[col])

    # check for time inversion
    for ind, row in time_disc_df.iterrows():
        if row["end"] < row["start"]:
            raise ValueError(f"Event {ind} has 'end time' before 'start time'")

    # add new rows for events with repetition
    new_rows_list = []
    for ind, rep in time_disc_df["repetition"].items():

        # yearly repetition
        if rep == "yearly":
            sta_dt = time_disc_df.loc[ind, "start"]
            end_dt = time_disc_df.loc[ind, "end"]

            # yearly repetition doesn't support start and end dates
            # on different years
            if sta_dt.year != end_dt.year:
                raise ValueError("Yearly repetition is not supported for start and end dates on different years.")

            # add new row for valid year
            for y in years_list:
                y1 = sta_dt.year
                if y1 != y:
                    new_rows_list.append(pd.Series([
                        sta_dt.replace(year=y),
                        end_dt.replace(year=y),
                        time_disc_df["t"].loc[ind],
                        None
                    ], index=time_disc_df.columns))

        elif rep is not None:
            raise ValueError(f"Repetition {rep} is not supported.")

    full_time_disc_df = pd.concat([
        time_disc_df,
        pd.DataFrame(new_rows_list)
    ], axis=0, ignore_index=True)

    # ajust one-day intervals
    for ind, row in full_time_disc_df.iterrows():
        # date formatted start times
            # if full day interval, ajust end date
            if row["start"] == row["end"]:
                full_time_disc_df.loc[ind, "end"] = row["end"] + timedelta(days=1)

    # create new t index
    tdiscr = pd.Series(nan_idx, index=ts.index)
    for t_idx in full_time_disc_df["t"].unique():
        # filter conditions for idx t and apply
        t_df = full_time_disc_df[full_time_disc_df["t"] == t_idx]
        for _, row in t_df.iterrows():
            tdiscr.loc[
                ((ts >= row["start"]) & (ts < row["end"]))
            ] = t_idx

    return tdiscr
